{/***************************************************************************
                            ComboboxCol.pas
                              ---------
                          A column combobox
                       the combobox for all purposes
                       2 revision : 11.12.2022
                    Initial Revision : Fri jan 3 2014

***************************************************************************/
*  This program is distributed in the hope that it will be useful,          *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                     *
*                                                                           *
*****************************************************************************

TComboboxCol for Typhoon
No Copyright (C) 2014  Claus Moeller
email: vig.hjaelp@gmail.com

This component is a Combobox with a stringgrid attached. The content of the combo is a stringgrid and accessed as such.
Rows and Columns can be added
Width, Alignment, Color, Font and visibility can be changed for each column.

Special thanks to H Page-Clark 2013 for Colcombo
and to Jesus Reyes Aguilar for his work on StringGrid where I took a lot from.
}

unit Comboboxcol;

interface

uses  Classes,
      Controls,
      Forms,
      Graphics,
      LCLProc,
      LCLType,
      LResources,
      StdCtrls,
      SysUtils,
      types,
      Grids,
      Dialogs;


type
  TComboboxcol = class;

  { TComboColumn }
  TComboColumn = class(TCollectionItem)
  private
    FWidth     : Integer;
    FAlignment : TAlignment;
    FFont      : TFont;
    FColor     : TColor;
    FVisible   : Boolean;
    procedure SetWidth(const Value: Integer);
    procedure SetAlignment(const Value:tAlignment);
    procedure SetFont(const Value:TFont);
    procedure SetColor(const Value:TColor);
    procedure SetVisible(const Value : Boolean);
  protected

  public
    constructor Create(aCollection:TCollection); override;
    destructor  Destroy; override;
    procedure   Assign(Source: TPersistent); override;
  published
    property Alignment : TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property Color     : TColor read FColor write SetColor;
    property Font      : TFont read FFont write SetFont;
    property Visible   : Boolean read FVisible write SetVisible default True;
    property Width     : Integer read FWidth write SetWidth default 100;
  end;

  { TComboColumns }
  TComboColumns = class(TCollection)
  private
    FOwner :  TComboBoxCol;
    function  GetItem(Index: Integer): TComboColumn;
    procedure SetItem(Index: Integer; const Value: TComboColumn);
  protected
    procedure Update(Item: TCollectionItem); override;
  public
    function  Add:TComboColumn;
    function  Insert(index: Integer): TComboColumn;
    property  Items[Index: Integer]: TComboColumn read GetItem write SetItem;
    constructor Create(AOwner:TComboboxCol);
    function  GetOwner: TPersistent; override;
  end;

  { TComboboxcol}
  TComboboxcol = class(TCustomComboBox)
  private
    FColumns       : TComboColumns;
    FGrid          : TCustomStringGrid;
    FEditColumn    : Integer;
    FColumnSpacing : Integer;
    FDelimiter     : AnsiChar;
    FShowColSeparators      : Boolean;
    FHighlightColor : TColor;
    function GetColumnCount : Integer;
    procedure DrawItemEvent(Control: TWinControl; Index: Integer; ARect: TRect; State: TOwnerDrawState);
    procedure TruncateText(Sender: TObject);
    procedure SetEditColumn(Value : Integer);
    function  GetEditText: string;
  protected
    procedure FontChanged(Sender: TObject); override;
//    class function GetControlClassDefaultSize: TSize; override;
    procedure GetItems; override;
    procedure SetItems(const Value: TStrings); override;
    procedure SetStyle(Val: TComboBoxStyle); override;
    function  GetCells(ACol, ARow: Integer): string;
    procedure SetCells(ACol, ARow: Integer; const AValue: string);
    function  ReadRowCount : Integer;
    procedure WriteRowCount(RowCount : Integer);
  public
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Clear; override;
    property    ColumnCount: integer read GetColumnCount;
    property    Cells[ACol, ARow: Integer]: string read GetCells write SetCells;
    property    EditText : string read GetEditText;
    function    IndexInColumnOf(aCol : Integer; s : string): Integer;
    function    IndexInRowOf(aRow : Integer; s : string): Integer;
    procedure   AddRow;
  published
    // new properties
    property HighlightColor : TColor read FHighlightColor write FHighlightColor;
    property RowCount : Integer read ReadRowCount write WriteRowCount;
    property Columns    : TComboColumns read FColumns write FColumns;
    property EditColumn : Integer read FEditColumn write SetEditColumn default 0;
    property ColumnSpacing: integer read FColumnSpacing write FColumnSpacing default 2;
    property ShowColSeparators: Boolean read FShowColSeparators write FShowColSeparators default False;
    // inherited comboBox properties
    property Align;
    property Anchors;
    property ArrowKeysTraverseList;
    property AutoComplete;
    property AutoCompleteText;
    property AutoDropDown;
    property AutoSelect;
    property AutoSize;
    property BidiMode;
    property BorderSpacing;
    property BorderStyle;
    property CharCase;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ItemHeight;
    property ItemIndex;
    property Items;
    property ItemWidth;
    property MaxLength;
    property OnChange;
    property OnChangeBounds;
    property OnClick;
    property OnCloseUp;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
//    property OnDrawItem;
    property OnEndDrag;
    property OnDropDown;
    property OnEditingDone;
    property OnEnter;
    property OnExit;
    property OnGetItems;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnSelect;
    property OnUTF8KeyPress;
    property ParentBidiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
//    property ReadOnly;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
  end;

procedure Register;

implementation

Const DummyChar : Char = '|';

procedure Register;
begin
  {$I comboboxcol_icon.ctrs}
  RegisterComponents('Misc',[TComboboxcol]);
end;


// TComboColumn

procedure TComboColumn.Assign(Source: TPersistent);
begin
 if Source is TComboColumn then
  begin
    Color := TComboColumn(source).Color;
    Width := TComboColumn(source).Width;
    Alignment := TComboColumn(source).Alignment;
    Font.Assign(TComboColumn(source).Font);
    Visible   := TComboColumn(source).Visible;
  end;
end;

constructor TComboColumn.Create(aCollection: TCollection);
begin
  inherited;
  FFont     := TFont.Create;
  FWidth    := 100;
  FColor    := clWindow;
  FVisible  := True;
  FAlignment:= taLeftJustify;
end;

destructor TComboColumn.Destroy;
begin
  FFont.Free;
  Inherited;
end;

procedure TComboColumn.SetAlignment(const Value: tAlignment);
begin
  FAlignment := Value;
  TComboColumns(Collection).FOwner.Invalidate;
end;

procedure TComboColumn.SetColor(const Value: TColor);
begin
  FColor := Value;
  TComboColumns(Collection).FOwner.Invalidate;
end;

procedure TComboColumn.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
  TComboColumns(Collection).FOwner.Invalidate;
end;

procedure TComboColumn.SetWidth(const Value: Integer);
begin
  FWidth := Value;
  TComboColumns(Collection).FOwner.Invalidate;
end;

procedure TComboColumn.SetVisible(const Value: Boolean);
begin
  FVisible := Value;
  TComboColumns(Collection).FOwner.Invalidate;
end;


// TComboColumns

function TComboColumns.Add: TComboColumn;
begin
  Result := TComboColumn(inherited Add);
//  FOwner.FGrid.ColCount:= Count;
end;

constructor TComboColumns.Create(AOwner: TComboboxCol);
begin
  inherited Create(TComboColumn);
  FOwner := AOwner;
end;

function TComboColumns.GetItem(Index: Integer): TComboColumn;
begin
  If (Index <= FOwner.ColumnCount) Then
    Result := TComboColumn(inherited Items[index]);
end;

function TComboColumns.GetOwner: tPersistent;
begin
  Result := FOwner;
end;

function TComboColumns.Insert(index: Integer): TComboColumn;
begin
  Result := TComboColumn(inherited Insert(index));
end;

procedure TComboColumns.SetItem(Index: Integer;
  const Value: TComboColumn);
begin
  inherited SetItem(Index, Value);
end;

procedure TComboColumns.Update(Item: TCollectionItem);
begin
  inherited;
end;


// TComboboxCol

function TComboboxCol.GetColumnCount: integer;
begin
  Result := FColumns.Count
end;

procedure TComboboxCol.DrawItemEvent(Control: TWinControl; Index: Integer;
  ARect: TRect; State: TOwnerDrawState);
var
  cb           : TCustomComboBox;
  Col          : Integer;
  ColDrawRect  : TRect;
  StartForText : LongInt;
  aDroppedDown : Boolean;
begin
  if (Index < 0) and not (Control is TCustomComboBox) then
    Exit;
  aDroppedDown := DroppedDown;
  cb := TCustomComboBox(Control);
  If (FColumns.Count > 0) and (aDroppedDown or (FEditColumn < 0)) Then
    Begin
      Col := 0;
      ColDrawRect := ARect;
      While Col < FColumns.Count Do
        begin
          case (odSelected in State) of
            False: begin
                     cb.Canvas.Brush.Color := FColumns.Items[Col].Color;
                   end;
            True:  begin
                     cb.Canvas.Brush.Color := FHighlightColor;
                   end;
          end; // End Case
          If FColumns.Items[Col].Visible Then
            Begin
              If Col < FColumns.Count - 1 Then
                ColDrawRect.Right := ColDrawRect.Left + FColumns.Items[Col].Width
              Else
                ColDrawRect.Right := ColDrawRect.Right + FColumns.Items[Col].Width;
              If FColumns.Items[Col].FFont <> cb.Canvas.Font Then
                Begin
                  cb.Canvas.Font    := FColumns.Items[Col].FFont;
                End;
              If FColumns.Items[Col].Alignment = taLeftJustify Then
                Begin
                  StartForText := ColDrawRect.Left + FColumnSpacing;
                End
              Else If FColumns.Items[Col].Alignment = taRightJustify Then
                Begin
                  StartForText := ColDrawRect.Right -
                    cb.Canvas.TextWidth(GetCells(Col,Index))- FColumnSpacing;
                End
              Else If FColumns.Items[Col].Alignment = taCenter Then
                Begin
                  StartForText := ColDrawRect.Left
                   + (FColumns.Items[Col].Width div 2) - (cb.Canvas.TextWidth(GetCells(Col,Index)) div 2);
                End;
              // Evt. draw rectangle
              cb.Canvas.FillRect(ColDrawRect);
              cb.Canvas.TextOut(StartForText, ColDrawRect.Top, GetCells(Col,Index));
              if FShowColSeparators and (Col > 0) then
                begin
                  cb.Canvas.MoveTo(ColDrawRect.Left, ColDrawRect.Top);
                  cb.Canvas.LineTo(ColDrawRect.Left, ColDrawRect.Bottom);
                end;
              ColDrawRect.Left := ColDrawRect.Right;
              Col := Col + 1;
            end
          Else
            Begin
              Col := Col + 1;
            End;
        end;
    end
  Else
    Begin
      If (FColumns.Count > 0) and (EditColumn < FColumns.Count) Then
        Begin
          ColDrawRect := ARect;
          ColDrawRect.Right := ColDrawRect.Left + FColumns.Items[FEditColumn].Width;
          If FColumns.Items[FEditColumn].FFont <> cb.Canvas.Font Then
            Begin
              cb.Canvas.Font    := FColumns.Items[FEditColumn].FFont;
            End;
          If FColumns.Items[FEditColumn].Alignment = taLeftJustify Then
            Begin
              StartForText := ColDrawRect.Left + FColumnSpacing;
            End
          Else If FColumns.Items[FEditColumn].Alignment = taRightJustify Then
            Begin
              StartForText := ColDrawRect.Right -
                cb.Canvas.TextWidth(GetCells(FEditColumn,Index))- FColumnSpacing;
            End
          Else If FColumns.Items[FEditColumn].Alignment = taCenter Then
            Begin
              StartForText := ColDrawRect.Left
               + (FColumns.Items[FEditColumn].Width div 2) - (cb.Canvas.TextWidth(GetCells(FEditColumn,Index)) div 2);
            End;
          cb.Canvas.FillRect(ColDrawRect);
          cb.Canvas.TextOut(ColumnSpacing+StartForText, ColDrawRect.Top, GetCells(FEditColumn,Index));   //cb.Items[cb.ItemIndex]);
        end
      Else
        Exit;
    end;
end;


procedure TComboboxCol.FontChanged(Sender: TObject);
begin
  inherited FontChanged(Sender);
end;

procedure TComboboxCol.GetItems;
begin
  inherited GetItems;
end;

procedure TComboboxCol.SetItems(const Value: TStrings);
begin
  inherited SetItems(Value);
end;

procedure TComboboxCol.SetStyle(Val: TComboBoxStyle);
begin
  if Val=csOwnerDrawFixed then
    inherited SetStyle(Val);
end;

procedure TComboboxCol.SetEditColumn(Value : Integer);
begin
  FEditColumn := Value;
end;

constructor TComboboxCol.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  SetStyle(csOwnerDrawFixed);
  FColumns          := TComboColumns.Create(Self);
  FGrid             := TCustomStringGrid.Create(Self);
//  FColumns.Add;
  FEditColumn       := 0;
  FDelimiter        := #9;
  FColumnSpacing    := 2;
  FHighlightColor   := clHighLight;
  FShowColSeparators:= False;
  OnDrawItem        := @DrawItemEvent;
  OnExit            := @TruncateText;
end;

destructor TComboboxCol.Destroy;
begin
  FreeThenNil(FGrid);
  FColumns.Free;
  inherited Destroy;
end;

function TComboboxCol.Getcells(aCol, aRow: Integer): string;
begin
  If FColumns.Count > 0 Then
    Begin
      Result := FGrid.Cells[aCol,aRow];
    end;
end;

procedure TComboboxCol.Setcells(aCol, aRow: Integer; const Avalue: string);
begin
  If FColumns.Count > 0 Then
    Begin
      FGrid.ColCount := FColumns.Count;
      FGrid.RowCount := Self.Items.Count;
      // Cells
      if (aRow >= FGrid.RowCount) then raise Exception.Create('Row index out of range');
      if (aCol >= FGrid.ColCount) then raise Exception.Create('Col index out of range');
      FGrid.Cells[aCol,aRow] := Avalue;
      If aValue = '' Then
        Begin
          Self.Items[aRow] := DummyChar;
        end
      Else
        Begin
          If Self.Items.count > 0 Then
            Begin
              If FEditColumn >= 0 Then
                Self.Items[aRow] := FGrid.Cells[FEditColumn,aRow]
              Else
                Self.Items[aRow] := FGrid.Cells[0,aRow]
            end;
        End;
    end;
end;

procedure TComboboxCol.AddRow;
begin
  If FColumns.Count > 0 Then
    Begin
      FGrid.RowCount := FGrid.RowCount + 1;
      Self.Items.Add(DummyChar);
    end
end;

function TComboboxCol.IndexInColumnOf(acol : Integer; s : string): Integer;
Begin
  if (aCol >= FGrid.ColCount) then raise Exception.Create('IndexColumnOf: col out of range');
  Result := FGrid.Cols[aCol].IndexOf(s);
end;

function TComboboxCol.IndexInRowOf(aRow : Integer; s : string): Integer;
Begin
  if (aRow >= FGrid.RowCount) then raise Exception.Create('IndexinRowOf: row out of range');
  Result := FGrid.Rows[aRow].IndexOf(s);
end;

procedure TComboboxCol.TruncateText(Sender: TObject);
begin
  Text := EditText;
end;

function TComboboxCol.GetEditText : string;
begin
  if (EditColumn >= 0) and (ItemIndex >= 0) then
    Begin
      If FColumns.Count > 0 Then
        Begin
          if (EditColumn >= FGrid.ColCount) then raise Exception.Create('EditColumn out of range');
          Result := Items[ItemIndex];
        end
    end
  else
    Begin
      If ItemIndex >= 0 Then
        Result := Items[ItemIndex];
    end;
end;

Procedure TComboboxCol.Clear;
Begin
  Inherited;
  FGrid.Clear;
  FColumns.Clear;
end;

Function TComboboxCol.ReadRowCount : Integer;
Begin
  Result := FGrid.RowCount;
end;

Procedure TComboboxCol.WriteRowCount(RowCount: Integer);
Begin
  FGrid.RowCount := RowCount;
end;

initialization
end.

